@@ -25,6 +25,7 @@
 #include <config.h>
 #endif
 
+#include <stdlib.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -44,6 +45,18 @@
 
 #include "glib-helper.h"
 
+/* Number of seconds to keep a sdp_session_t in the cache */
+#define CACHE_TIMEOUT 2
+
+struct cached_sdp_session {
+	bdaddr_t src;
+	bdaddr_t dst;
+	sdp_session_t *session;
+	guint timer;
+};
+
+static GSList *cached_sdp_sessions = NULL;
+
 typedef int (*resolver_t) (int fd, char *src, char *dst);
 typedef BtIOError (*connect_t) (BtIO *io, BtIOFunc func);
 typedef BtIOError (*listen_t) (BtIO *io, BtIOFunc func);
@@ -55,6 +68,62 @@ struct hci_cmd_data {
 	gpointer		caller_data;
 };
 
+static gboolean cached_session_expired(gpointer user_data)
+{
+	struct cached_sdp_session *cached = user_data;
+
+	cached_sdp_sessions = g_slist_remove(cached_sdp_sessions, cached);
+
+	sdp_close(cached->session);
+
+	g_free(cached);
+
+	return FALSE;
+}
+
+static sdp_session_t *get_sdp_session(const bdaddr_t *src, const bdaddr_t *dst)
+{
+	GSList *l;
+
+	for (l = cached_sdp_sessions; l != NULL; l = l->next) {
+		struct cached_sdp_session *c = l->data;
+		sdp_session_t *session;
+
+		if (bacmp(&c->src, src) || bacmp(&c->dst, dst))
+			continue;
+
+		g_source_remove(c->timer);
+
+		session = c->session;
+
+		cached_sdp_sessions = g_slist_remove(cached_sdp_sessions, c);
+		g_free(c);
+
+		return session;
+	}
+
+	return sdp_connect(src, dst, SDP_NON_BLOCKING);
+}
+
+static void cache_sdp_session(bdaddr_t *src, bdaddr_t *dst,
+				sdp_session_t *session)
+{
+	struct cached_sdp_session *cached;
+
+	cached = g_new0(struct cached_sdp_session, 1);
+
+	bacpy(&cached->src, src);
+	bacpy(&cached->dst, dst);
+
+	cached->session = session;
+
+	cached_sdp_sessions = g_slist_append(cached_sdp_sessions, cached);
+
+	cached->timer = g_timeout_add_seconds(CACHE_TIMEOUT,
+						cached_session_expired,
+						cached);
+}
+
 int set_nonblocking(int fd)
 {
 	long arg;
@@ -161,7 +230,7 @@ static void search_completed_cb(uint8_t type, uint16_t status,
 	} while (scanned < size && bytesleft > 0);
 
 done:
-	sdp_close(ctxt->session);
+	cache_sdp_session(&ctxt->src, &ctxt->dst, ctxt->session);
 
 	if (ctxt->cb)
 		ctxt->cb(recs, err, ctxt->user_data);
@@ -264,7 +333,7 @@ static int create_search_context(struct search_context **ctxt,
 	if (!ctxt)
 		return -EINVAL;
 
-	s = sdp_connect(src, dst, SDP_NON_BLOCKING);
+	s = get_sdp_session(src, dst);
 	if (!s)
 		return -errno;
 
@@ -418,7 +487,7 @@ static struct {
 	{ }
 };
 
-uint16_t bt_string2class(const char *pattern)
+uint16_t bt_name2class(const char *pattern)
 {
 	int i;
 
@@ -430,16 +499,51 @@ uint16_t bt_string2class(const char *pattern)
 	return 0;
 }
 
+static inline gboolean is_uuid128(const char *string)
+{
+	return (strlen(string) == 36 &&
+			string[8] == '-' &&
+			string[13] == '-' &&
+			string[18] == '-' &&
+			string[23] == '-');
+}
+
+char *bt_name2string(const char *pattern)
+{
+	uuid_t uuid;
+	uint16_t uuid16;
+	int i;
+
+	/* UUID 128 string format */
+	if (is_uuid128(pattern))
+		return g_strdup(pattern);
+
+	/* Friendly service name format */
+	uuid16 = bt_name2class(pattern);
+	if (uuid16)
+		goto proceed;
+
+	/* HEX format */
+	uuid16 = strtol(pattern, NULL, 16);
+	for (i = 0; bt_services[i].class; i++) {
+		if (bt_services[i].class == uuid16)
+			goto proceed;
+	}
+
+	return NULL;
+
+proceed:
+	sdp_uuid16_create(&uuid, uuid16);
+
+	return bt_uuid2string(&uuid);
+}
+
 int bt_string2uuid(uuid_t *uuid, const char *string)
 {
 	uint32_t data0, data4;
 	uint16_t data1, data2, data3, data5;
 
-	if (strlen(string) == 36 &&
-			string[8] == '-' &&
-			string[13] == '-' &&
-			string[18] == '-' &&
-			string[23] == '-' &&
+	if (is_uuid128(string) &&
 			sscanf(string, "%08x-%04hx-%04hx-%04hx-%08x%04hx",
 				&data0, &data1, &data2, &data3, &data4, &data5) == 6) {
 		uint8_t val[16];
@@ -462,7 +566,7 @@ int bt_string2uuid(uuid_t *uuid, const char *string)
 
 		return 0;
 	} else {
-		uint16_t class = bt_string2class(string);
+		uint16_t class = bt_name2class(string);
 		if (class) {
 			sdp_uuid16_create(uuid, class);
 			return 0;
@@ -775,7 +879,7 @@ static BtIOError sco_connect(BtIO *io, BtIOFunc func)
 {
 	struct io_context *io_ctxt = io->io_ctxt;
 	struct sockaddr_sco addr;
-	int sk, err;
+	int err;
 
 	io_ctxt->func = func;
 
@@ -790,7 +894,7 @@ static BtIOError sco_connect(BtIO *io, BtIOFunc func)
 	err = transport_connect(io, (struct sockaddr *) &addr,
 				sizeof(addr));
 	if (err < 0) {
-		close(sk);
+		close(io_ctxt->fd);
 		return BT_IO_FAILED;
 	}
 
@@ -844,18 +948,18 @@ static BtIOError l2cap_listen(BtIO *io, BtIOFunc func)
 {
 	struct io_context *io_ctxt = io->io_ctxt;
 	struct sockaddr_l2 addr;
-	BtIOError err;
+	int err;
 
 	io_ctxt->func = func;
 
 	err = l2cap_bind(io_ctxt, io->src, io->psm, io->mtu, io->flags, &addr);
 	if (err < 0)
-		return err;
+		return BT_IO_FAILED;
 
 	err = transport_listen(io);
 	if (err < 0) {
 		close(io_ctxt->fd);
-		return err;
+		return BT_IO_FAILED;
 	}
 
 	return BT_IO_SUCCESS;
@@ -865,13 +969,13 @@ static BtIOError l2cap_connect(BtIO *io, BtIOFunc func)
 {
 	struct io_context *io_ctxt = io->io_ctxt;
 	struct sockaddr_l2 l2a;
-	BtIOError err;
+	int err;
 
 	io_ctxt->func = func;
 
 	err = l2cap_bind(io_ctxt, io->src, 0, io->mtu, 0, &l2a);
 	if (err < 0)
-		return err;
+		return BT_IO_FAILED;
 
 	memset(&l2a, 0, sizeof(l2a));
 	l2a.l2_family = AF_BLUETOOTH;
@@ -882,7 +986,7 @@ static BtIOError l2cap_connect(BtIO *io, BtIOFunc func)
 				sizeof(l2a));
 	if (err < 0) {
 		close(io_ctxt->fd);
-		return err;
+		return BT_IO_FAILED;
 	}
 
 	return BT_IO_SUCCESS;
@@ -892,7 +996,7 @@ static BtIOError rfcomm_bind(struct io_context *io_ctxt, const char *address,
 				uint8_t channel, uint32_t flags,
 				struct sockaddr_rc *addr)
 {
-	BtIOError err;
+	int err;
 
 	io_ctxt->fd = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
 	if (io_ctxt->fd < 0)
@@ -928,18 +1032,18 @@ static BtIOError rfcomm_listen(BtIO *io, BtIOFunc func)
 	struct io_context *io_ctxt = io->io_ctxt;
 	struct sockaddr_rc addr;
 	socklen_t sa_len;
-	BtIOError err;
+	int err;
 
 	io_ctxt->func = func;
 
 	err = rfcomm_bind(io_ctxt, io->src, io->channel, io->flags, &addr);
 	if (err < 0)
-		return err;
+		return BT_IO_FAILED;
 
 	err = transport_listen(io);
 	if (err < 0) {
 		close(io_ctxt->fd);
-		return err;
+		return BT_IO_FAILED;
 	}
 
 	sa_len = sizeof(struct sockaddr_rc);
@@ -947,7 +1051,7 @@ static BtIOError rfcomm_listen(BtIO *io, BtIOFunc func)
 	if (getsockname(io_ctxt->fd, (struct sockaddr *) &addr, &sa_len) < 0) {
 		err = -errno;
 		close(io_ctxt->fd);
-		return err;
+		return BT_IO_FAILED;
 	}
 
 	io->channel = addr.rc_channel;
@@ -959,13 +1063,13 @@ static BtIOError rfcomm_connect(BtIO *io, BtIOFunc func)
 {
 	struct io_context *io_ctxt = io->io_ctxt;
 	struct sockaddr_rc addr;
-	BtIOError err;
+	int err;
 
 	io_ctxt->func = func;
 
 	err = rfcomm_bind(io_ctxt, io->src, 0, 0, &addr);
 	if (err < 0)
-		return err;
+		return BT_IO_FAILED;
 
 	memset(&addr, 0, sizeof(addr));
 	addr.rc_family = AF_BLUETOOTH;
@@ -976,7 +1080,7 @@ static BtIOError rfcomm_connect(BtIO *io, BtIOFunc func)
 				sizeof(addr));
 	if (err < 0) {
 		close(io_ctxt->fd);
-		return err;
+		return BT_IO_FAILED;
 	}
 
 	return BT_IO_SUCCESS;
@@ -986,18 +1090,18 @@ static BtIOError sco_listen(BtIO *io, BtIOFunc func)
 {
 	struct io_context *io_ctxt = io->io_ctxt;
 	struct sockaddr_sco addr;
-	BtIOError err;
+	int err;
 
 	io_ctxt->func = func;
 
 	err = sco_bind(io_ctxt, io->src, io->mtu, &addr);
 	if (err < 0)
-		return err;
+		return BT_IO_FAILED;
 
 	err = transport_listen(io);
 	if (err < 0) {
 		close(io_ctxt->fd);
-		return err;
+		return BT_IO_FAILED;
 	}
 
 	return BT_IO_SUCCESS;
@@ -1206,7 +1310,7 @@ static void io_context_cleanup(struct io_context *io_ctxt)
 	g_free(io_ctxt);
 }
 
-GIOChannel *rfcomm_listen_internal(const bdaddr_t *src, uint8_t *channel,
+static GIOChannel *rfcomm_listen_internal(const bdaddr_t *src, uint8_t *channel,
 			uint32_t flags, bt_io_callback_t cb, void *user_data)
 {
 	BtIO *io;
