@@ -247,9 +247,10 @@ static DBusMessage *list_adapters(DBusConnection *conn,
 	for (l = adapters; l; l = l->next) {
 		struct btd_adapter *adapter = l->data;
 		struct hci_dev_info di;
-		dev_id = adapter_get_dev_id(adapter);
 		const gchar *path = adapter_get_path(adapter);
 
+		dev_id = adapter_get_dev_id(adapter);
+
 		if (hci_devinfo(dev_id, &di) < 0)
 			continue;
 
@@ -326,8 +327,53 @@ dbus_bool_t manager_init(DBusConnection *conn, const char *path)
 			NULL, NULL, NULL);
 }
 
+static void manager_update_adapters(void)
+{
+	GSList *list;
+	char **array;
+	int i;
+
+	array = g_new0(char *, g_slist_length(adapters) + 1);
+	for (i = 0, list = adapters; list; list = list->next, i++) {
+		struct btd_adapter *adapter = list->data;
+		array[i] = (char *) adapter_get_path(adapter);
+	}
+
+	emit_array_property_changed(connection, "/",
+					MANAGER_INTERFACE, "Adapters",
+					DBUS_TYPE_OBJECT_PATH, &array);
+
+	g_free(array);
+}
+
+static void manager_remove_adapter(struct btd_adapter *adapter)
+{
+	uint16_t dev_id = adapter_get_dev_id(adapter);
+	const gchar *path = adapter_get_path(adapter);
+
+	adapters = g_slist_remove(adapters, adapter);
+
+	manager_update_adapters();
+
+	if (default_adapter_id == dev_id || default_adapter_id < 0) {
+		int new_default = hci_get_route(NULL);
+
+		manager_set_default_adapter(new_default);
+	}
+
+	g_dbus_emit_signal(connection, "/",
+			MANAGER_INTERFACE, "AdapterRemoved",
+			DBUS_TYPE_OBJECT_PATH, &path,
+			DBUS_TYPE_INVALID);
+
+	adapter_remove(adapter);
+}
+
 void manager_cleanup(DBusConnection *conn, const char *path)
 {
+	g_slist_foreach(adapters, (GFunc) manager_remove_adapter, NULL);
+	g_slist_free(adapters);
+
 	g_dbus_unregister_interface(conn, "/", MANAGER_INTERFACE);
 }
 
@@ -407,42 +453,25 @@ static void manager_add_adapter(struct btd_adapter *adapter)
 {
 	const gchar *path = adapter_get_path(adapter);
 
+	adapters = g_slist_append(adapters, adapter);
+
 	g_dbus_emit_signal(connection, "/",
 			MANAGER_INTERFACE, "AdapterAdded",
 			DBUS_TYPE_OBJECT_PATH, &path,
 			DBUS_TYPE_INVALID);
 
-	adapters = g_slist_append(adapters, adapter);
-
 	if (default_adapter_id < 0) {
 		int new_default = hci_get_route(NULL);
 
 		manager_set_default_adapter(new_default);
 	}
+
+	manager_update_adapters();
 }
 
-static void manager_remove_adapter(struct btd_adapter *adapter)
+int manager_register_adapter(int id, gboolean devup)
 {
-	uint16_t dev_id = adapter_get_dev_id(adapter);
-	const gchar *path = adapter_get_path(adapter);
-
-	g_dbus_emit_signal(connection, "/",
-			MANAGER_INTERFACE, "AdapterRemoved",
-			DBUS_TYPE_OBJECT_PATH, &path,
-			DBUS_TYPE_INVALID);
-
-	if (default_adapter_id == dev_id || default_adapter_id < 0) {
-		int new_default = hci_get_route(NULL);
-
-		manager_set_default_adapter(new_default);
-	}
-
-	adapters = g_slist_remove(adapters, adapter);
-}
-
-int manager_register_adapter(int id)
-{
-	struct btd_adapter *adapter = adapter_create(connection, id);
+	struct btd_adapter *adapter = adapter_create(connection, id, devup);
 
 	if (!adapter)
 		return -1;
@@ -467,8 +496,6 @@ int manager_unregister_adapter(int id)
 
 	manager_remove_adapter(adapter);
 
-	adapter_remove(adapter);
-
 	return 0;
 }
 
